'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var stream = require('readable-stream');
var stream__default = _interopDefault(stream);
var universalify = require('universalify');
var util = require('util');
var util__default = _interopDefault(util);

function isBufferish (chunk) {
  return Buffer.isBuffer(chunk) || chunk instanceof Uint8Array
}

function guessTarget ({ chunk }) {
  if (isBufferish(chunk)) {
    return 'buffer'
  } else if (typeof chunk === 'string') {
    return 'string'
  } else {
    return 'array'
  }
}

function asBuffer (items) {
  return Buffer.concat(
    items.map(
      ({ chunk, encoding }) => isBufferish(chunk)
        ? chunk
        : Buffer.from(chunk, encoding)
    )
  )
}

function toString ({ chunk, encoding }) {
  if (typeof chunk === 'string') {
    return chunk
  } else if (Buffer.isBuffer(chunk)) {
    return chunk.toString(encoding)
  } else if (chunk instanceof Uint8Array) {
    return Buffer.from(chunk).toString(encoding)
  } else {
    return chunk.toString()
  }
}

function asString (items) {
  return items.map(toString).join('')
}

function compileItems (items, target) {
  if (target === undefined && items.length > 0) {
    target = guessTarget(items[0]);
  }
  if (target === 'buffer') {
    return asBuffer(items)
  } else if (target === 'string') {
    return asString(items)
  } else {
    return items.map(item => item.chunk)
  }
}

function collect (target) {
  const items = [];
  return new stream.Transform({
    objectMode: true,
    transform (chunk, encoding, callback) {
      items.push({ chunk, encoding });
      callback();
    },
    flush (callback) {
      let err = null;
      let res;
      try {
        res = compileItems(items, target);
      } catch (e) {
        err = e;
      }
      callback(err, res);
    }
  })
}

function isFunction (value) {
  return typeof value === 'function'
}

function isPlainObject (value) {
  if (typeof value !== 'object' || value === null) {
    return false
  }
  if (Object.getPrototypeOf(value) === null) {
    return true
  }
  let proto = value;
  while (Object.getPrototypeOf(proto) !== null) {
    proto = Object.getPrototypeOf(proto);
  }
  return Object.getPrototypeOf(value) === proto
}

function isPromise (value) {
  return util__default.types.isPromise(value)
}

function once (callback) {
  let called = false;
  return function (err, res) {
    if (called) return
    called = true;
    callback(err, res);
  }
}

function makeConcurrentMethods (concurrency, _write, _final) {
  if (!Number.isInteger(concurrency) || concurrency <= 0) {
    throw new TypeError('Concurrency must be a positive integer')
  }

  // Active jobs counter
  let jobs = 0;

  // First encountered error
  let error;

  // Require more chunks callback
  let next;

  // End stream callback
  let close;

  function concurrentWrite (chunk, encoding, callback) {
    if (!error) {
      jobStart(this, chunk, encoding);
    }

    if (!error && jobs < concurrency) {
      // Keep requiring more chunks
      callback();
    } else {
      // Save the current callback for jobs completion
      next = callback;
    }
  }

  function jobStart (stream, chunk, encoding) {
    jobs++;
    _write.call(
      stream,
      chunk,
      encoding,
      once((err, data) => jobEnd(stream, err, data))
    );
  }

  function jobEnd (stream, err, data) {
    jobs--;
    error = error || err;

    if (!error && data !== undefined && stream.push) {
      stream.push(data);
    }

    if (jobs <= 0) {
      const callback = close || next;
      if (callback) {
        callback(error);
      } else if (error) {
        stream.emit('error', error);
      }
    } else if (!error && next) {
      const callback = next;
      next = undefined;
      callback();
    }
  }

  function concurrentFinal (callback) {
    // Build close callback
    close = err => {
      if (err || !_final) {
        callback(err);
      } else {
        _final.call(this, callback);
      }
    };

    if (jobs <= 0) {
      close();
    }
  }

  return [concurrentWrite, concurrentFinal]
}

function patchWritable (stream, concurrency) {
  const [_write, _final] = makeConcurrentMethods(
    concurrency,
    stream._write,
    stream._final
  );
  stream._write = _write;
  stream._final = _final;
}

function patchTransform (stream, concurrency) {
  const [_transform, _flush] = makeConcurrentMethods(
    concurrency,
    stream._transform,
    stream._flush
  );
  stream._transform = _transform;
  stream._flush = _flush;
}

function handlePromise (fn) {
  if (!isFunction(fn)) {
    return fn
  }
  return Object.defineProperty(
    function (...args) {
      const callback = args[args.length - 1];
      const out = fn.call(this, ...args);
      if (isPromise(out)) {
        out
          .then(result => callback(null, result))
          .catch(callback);
      }
    },
    'name',
    { value: fn.name }
  )
}

function createReadMethod (asyncRead) {
  let reading = false;
  return function _read (size) {
    if (reading) {
      return
    }
    reading = true;

    const self = this;
    function callback (err, data) {
      if (err) {
        process.nextTick(() => self.emit('error', err));
        return
      }

      if (data !== undefined) {
        self.push(data);
      }

      const state = self._readableState || {};

      if (data === null || !!state.ended) {
        return
      }

      reading = false;

      if (state.length < state.highWaterMark) {
        setImmediate(() => self._read(size));
      }
    }

    asyncRead.call(this, size, once(callback));
  }
}

function Duplex (options) {
  if (!(this instanceof Duplex)) {
    return new Duplex(options)
  }
  stream__default.Duplex.call(this, options);
  if (options && options.asyncRead) {
    this._asyncRead = options.asyncRead;
  }
  this._asyncRead = handlePromise(this._asyncRead);
  this._write = handlePromise(this._write);
  this._writev = handlePromise(this._writev);
  this._final = handlePromise(this._final);
  this._destroy = handlePromise(this._destroy);
  if (this._asyncRead) {
    this._read = createReadMethod(this._asyncRead);
  }
  if (options && options.concurrency > 0) {
    patchWritable(this, options.concurrency);
  }
}

util.inherits(Duplex, stream__default.Duplex);

function isRequest (stream) {
  return !!stream.setHeader && isFunction(stream.abort)
}

function destroyStream (stream, err) {
  if (isRequest(stream)) {
    stream.abort();
  } else if (isFunction(stream.destroy)) {
    stream.destroy(err);
  }
}

function noSource () {
  this.push(null);
}

function noTarget (chunk, encoding, callback) {
  callback(null);
}

function duplexify (options, readable, writable) {
  if (!isPlainObject(options)) {
    writable = readable;
    readable = options;
    options = {};
  }

  let listener;

  function read () {
    if (listener) {
      readable.resume();
      return
    }

    listener = data => {
      if (!this.push(data)) {
        readable.pause();
      }
    };

    stream.finished(readable, { writable: false }, err => {
      readable.removeListener('data', listener);

      if (err) {
        this.emit('error', err);
      } else {
        this.push(null);
      }
    });

    readable.addListener('data', listener);
  }

  function write (chunk, encoding, callback) {
    writable.write(chunk, encoding, callback);
  }

  function final (callback) {
    writable.end(callback);
  }

  function destroy (err, callback) {
    if (readable) {
      destroyStream(readable, err);
    }
    if (writable) {
      destroyStream(writable, err);
    }
    callback(err);
  }

  return new stream.Duplex({
    readableHighWaterMark: readable ? readable.readableHighWaterMark : undefined,
    readableObjectMode: readable ? readable.readableObjectMode : undefined,
    writableHighWaterMark: writable ? writable.writableHighWaterMark : undefined,
    writableObjectMode: writable ? writable.writableObjectMode : undefined,
    ...options,
    read: readable ? read : noSource,
    write: writable ? write : noTarget,
    writev: undefined,
    final: writable ? final : undefined,
    destroy
  })
}

function isReadable (value) {
  return value instanceof stream.Stream && isFunction(value.push)
}

function isWritable (value) {
  return value instanceof stream.Stream && isFunction(value.write)
}

function isStream (value) {
  return isReadable(value) || isWritable(value)
}

function isDuplex (value) {
  return isReadable(value) && isWritable(value)
}

function isReadableStrictly (value) {
  return isReadable(value) && !isWritable(value)
}

function isWritableStrictly (value) {
  return isWritable(value) && !isReadable(value)
}

function Readable (options) {
  if (!(this instanceof Readable)) {
    return new Readable(options)
  }
  stream__default.Readable.call(this, options);
  if (options && options.asyncRead) {
    this._asyncRead = options.asyncRead;
  }
  this._asyncRead = handlePromise(this._asyncRead);
  this._destroy = handlePromise(this._destroy);
  if (this._asyncRead) {
    this._read = createReadMethod(this._asyncRead);
  }
}

util.inherits(Readable, stream__default.Readable);

Readable.from = stream__default.Readable.from;

function readify (...streams) {
  const options = isPlainObject(streams[0]) ? streams.shift() : {};

  if (streams.length <= 0) {
    return new stream.Readable(options)
  } else if (streams.length === 1) {
    return streams[0]
  }

  let next;
  let tail;

  return new stream.Readable({
    ...options,
    read () {
      if (next) {
        const callback = next;
        next = undefined;
        callback();
      }
      if (tail) {
        return
      }

      const self = this;

      tail = new stream.Writable({
        highWaterMark: options.highWaterMark || options.readableHighWaterMark,
        objectMode: options.objectMode || options.readableObjectMode,
        write (chunk, encoding, callback) {
          if (self.push(chunk, encoding)) {
            callback();
          } else {
            next = callback;
          }
        }
      });

      stream.pipeline(...streams, tail, err => {
        process.nextTick(() => {
          if (err) {
            self.emit('error', err);
          } else {
            self.push(null);
          }
        });
      });
    },
    destroy (err, callback) {
      if (tail) {
        tail.destroy(err);
      }
      callback(err);
    }
  })
}

function subscribeStream (...args) {
  const end = args.pop();
  let result = null;
  stream.pipeline(
    ...args,
    new stream.Writable({
      objectMode: true,
      write (chunk, encoding, callback) {
        result = chunk;
        callback();
      }
    }),
    err => end(err, result)
  );
}

const subscribe = universalify.fromCallback(subscribeStream);

function Transform (options) {
  if (!(this instanceof Transform)) {
    return new Transform(options)
  }
  stream__default.Transform.call(this, options);
  this._transform = handlePromise(this._transform);
  this._flush = handlePromise(this._flush);
  this._destroy = handlePromise(this._destroy);
  if (options && options.concurrency > 0) {
    patchTransform(this, options.concurrency);
  }
}

util.inherits(Transform, stream__default.Transform);

function Writable (options) {
  if (!(this instanceof Writable)) {
    return new Writable(options)
  }
  stream__default.Writable.call(this, options);
  this._write = handlePromise(this._write);
  this._writev = handlePromise(this._writev);
  this._final = handlePromise(this._final);
  this._destroy = handlePromise(this._destroy);
  if (options && options.concurrency > 0) {
    patchWritable(this, options.concurrency);
  }
}

util.inherits(Writable, stream__default.Writable);

function writify (...streams) {
  const options = isPlainObject(streams[0]) ? streams.shift() : {};

  if (streams.length <= 0) {
    return new stream.Writable(options)
  } else if (streams.length === 1) {
    return streams[0]
  }

  let error;
  let head;
  let next;

  function handle (err) {
    error = err || error;
    if (next) {
      const callback = next;
      next = undefined;
      callback(error);
    }
  }

  return new stream.Writable({
    ...options,
    write (chunk, encoding, callback) {
      next = callback;

      if (!head) {
        head = streams[0];
        stream.pipeline(...streams, handle);
      }

      if (error) {
        handle(error);
        return
      }

      if (head.write(chunk, encoding)) {
        handle();
      } else {
        head.once('drain', handle);
      }
    },
    final (callback) {
      next = callback;

      if (error) {
        handle(error);
      } else {
        head.end();
      }
    },
    destroy (err, callback) {
      if (head) {
        destroyStream(head);
      }
      callback(err);
    }
  })
}

const finished = universalify.fromCallback(stream__default.finished);
const pipeline = universalify.fromCallback(stream__default.pipeline);

Object.defineProperty(exports, 'PassThrough', {
  enumerable: true,
  get: function () {
    return stream.PassThrough;
  }
});
Object.defineProperty(exports, 'Stream', {
  enumerable: true,
  get: function () {
    return stream.Stream;
  }
});
exports.Duplex = Duplex;
exports.Readable = Readable;
exports.Transform = Transform;
exports.Writable = Writable;
exports.collect = collect;
exports.duplexify = duplexify;
exports.finished = finished;
exports.isDuplex = isDuplex;
exports.isReadable = isReadable;
exports.isReadableStrictly = isReadableStrictly;
exports.isStream = isStream;
exports.isWritable = isWritable;
exports.isWritableStrictly = isWritableStrictly;
exports.pipeline = pipeline;
exports.readify = readify;
exports.subscribe = subscribe;
exports.writify = writify;
